/* Copyright 2015, 2016 OpenMarket Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * This is a manual index of history_visibility content of state events,
 * so that we can join on them in SELECT statements.
 */
CREATE TABLE IF NOT EXISTS history_visibility(
    event_id TEXT NOT NULL,
    room_id TEXT NOT NULL,
    history_visibility TEXT NOT NULL,
    UNIQUE (event_id)
);
